
def remove(images, docker, requester, image_tag):
    image = images.get_user_unmounted_image_from_tag(requester, image_tag)
    if image:   # otherwise issue is already reported
        docker.rmi(image.fullname)
        images.remove(image.fullname)




from subprocess import check_output

class Filesystem(object):
    def __init__(self, cmd_pattern):
        self.cmd_pattern = cmd_pattern
    def wrap_cmd(self, cmd):
        cmd_args = cmd.split()
        return self.cmd_pattern % dict(
            prog = cmd_args[0],
            prog_args = ' '.join(cmd_args[1:])
        ) + ' 2>/dev/null || true'
    def run_cmd(self, cmd):
        return check_output(self.wrap_cmd(cmd), shell=True)
    def ping(self):
        return self.run_cmd('echo ok').strip() == 'ok'
    def get_file_type(self, path):
        if len(self.run_cmd('find %s' % path)) == 0:
            return None
        for ftype in [ 'f', 'd' ]:
            check_cmd = 'find %(path)s -type %(ftype)s -maxdepth 0 -printf %(ftype)s' % \
                dict(
                    path = path,
                    ftype = ftype
                )
            result = self.run_cmd(check_cmd)
            if len(result) > 0:
                return result
        return 'o'  # other



from collections import namedtuple
import cPickle as pickle
import re

COLUMNATE_SPACING = 2

PARAGRAPH_FORMATING = """\

\033[1m\
%(title)s
\033[0m\

%(content)s

%(footnote)s"""

# use the following like this:
#
# with AutoCleaner(<obj>) as <var>:
#     ... work_with <var> ...
#
# <obj> must provide a method cleanup()
# that will be called automatically when leaving 
# the with construct. 

class AutoCleaner(object):
    def __init__(self, obj):
        self.obj = obj
    def __enter__(self):
        return self.obj
    def __exit__(self, t, value, traceback):
        self.obj.cleanup()
        self.obj = None

def as_string(item):
    if item == None:
        return ''
    else:
        return str(item)

def columnate(tabular_data, header = None):
    # stringify all (and work on a copy)
    tabular_data_copy = [ [as_string(s) for s in i] for i in tabular_data ]
    # if header is specified, add it
    # and replace underscores with spaces
    if header != None:
        tabular_data_copy.insert(0, [ i.replace('_', ' ') for i in header ])
    # compute the max length of elements in each column
    colwidths = [ max([ len(s) for s in i ]) for i in zip(*tabular_data_copy) ]
    # if header, underline it 
    if header != None:
        tabular_data_copy.insert(1, [ '-' * w for w in colwidths ])
    # compute a format that should be applied to each record
    formating = "".join([ '%-' + str(w + COLUMNATE_SPACING) + 's' \
                            for w in colwidths ])
    # format and return
    return '\n'.join(formating % tuple(record) for record in tabular_data_copy)

def display_transient_label(stdout, label):
    stdout.write('\r' + label + ' ')
    stdout.flush()

def hide_transient_label(stdout, label):
    # override with space
    stdout.write('\r%s\r' % (' '*len(label)))
    stdout.flush()

def format_paragraph(title, content, footnote=None):
    if footnote:
        footnote += '\n\n'
    else:
        footnote = ''
    return PARAGRAPH_FORMATING % dict(
                                    title = title,
                                    content = content,
                                    footnote = footnote)


MAX_PRINTED_NODES = 10
CONJUGATE_REGEXP = r'\b(\w*)\((\w*)\)'

def format_sentence_about_nodes(sentence, nodes):
    """
    example 1:
        input: sentence = '% seems(seem) dead.', nodes = ['rpi0']
        output: 'Node rpi0 seems dead.'
    example 2:
        input: sentence = '% seems(seem) dead.', nodes = ['rpi0', 'rpi1', 'rpi2']
        output: 'Nodes rpi0, rpi1 and rpi2 seem dead.'
    (and if there are many nodes, an ellipsis is used.)
    """
    # conjugate if plural or singular
    if len(nodes) > 1:
        sentence = re.sub(CONJUGATE_REGEXP, r'\2', sentence)
    else:
        sentence = re.sub(CONJUGATE_REGEXP, r'\1', sentence)
    # designation of nodes
    sorted_nodes = sorted(nodes)
    if len(nodes) > MAX_PRINTED_NODES:
        s_nodes = 'Nodes %s, %s, ..., %s' % (sorted_nodes[0], sorted_nodes[1], sorted_nodes[-1])
    elif len(nodes) > 1:
        s_nodes = 'Nodes %s and %s' % (', '.join(sorted_nodes[:-1]), sorted_nodes[-1])
    else:
        s_nodes = 'Node %s' % tuple(nodes)
    # almost done
    return sentence % s_nodes

# are you sure you want to understand what follows? This is sorcery...
nt_index = 0
nt_classes = {}
def to_named_tuple(d):
    global nt_index
    code = pickle.dumps(sorted(d.keys()))
    if code not in nt_classes:
        nt_classes[code] = namedtuple('NamedTuple_%d' % nt_index, d.keys())
        nt_index += 1
    return nt_classes[code](**d)

def merge_named_tuples(nt1, nt2):
    d = nt1._asdict()
    d.update(nt2._asdict())
    return to_named_tuple(d)




import uuid, time
from walt.server.images.image import parse_image_fullname, validate_image_tag

# About terminology: See comment about it in image.py.
class ImageShellSessionStore(object):
    def __init__(self, docker, images):
        self.docker = docker
        self.images = images
        self.sessions = set()
    def create_session(self, requester, image_tag):
        image = self.images.get_user_image_from_tag(requester, image_tag)
        if image:
            session = ImageShellSession(self, self.images,
                                    requester, image.fullname)
            self.sessions.add(session)
            return session
    def cleanup_session(self, session):
        session.cleanup()
        self.sessions.remove(session)
    def cleanup(self):
        for session in self.sessions:
            session.cleanup()
        self.sessions = {}

class ImageShellSession(object):
    NAME_OK             = 0
    NAME_NOT_OK         = 1
    NAME_NEEDS_CONFIRM  = 2
    exposed_NAME_OK             = NAME_OK
    exposed_NAME_NOT_OK         = NAME_NOT_OK
    exposed_NAME_NEEDS_CONFIRM  = NAME_NEEDS_CONFIRM
    def __init__(self, store, images, requester, image_fullname):
        self.store = store
        self.docker = store.docker
        self.images = images
        self.requester = requester
        self.image_fullname, dummy1, dummy2, dummy3, self.image_tag = \
            parse_image_fullname(image_fullname)
        self.new_image_tag = None
        self.container_name = str(uuid.uuid4())
        # expose methods to the RPyC client
        self.exposed___enter__ = self.__enter__
        self.exposed___exit__ = self.__exit__
        self.exposed_get_parameters = self.get_parameters
        self.exposed_get_default_new_name = self.get_default_new_name
        self.exposed_validate_new_name = self.validate_new_image_tag
        self.exposed_select_new_name = self.select_new_name
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.finalize()
    def get_parameters(self):
        # return an immutable object (a tuple, not a dict)
        # otherwise we will cause other RPyC calls
        return self.image_fullname, self.container_name
    def get_default_new_name(self):
        # default is to propose the same name
        # (and override the image if user confirms)
        return self.image_tag
    def select_new_name(self, new_image_tag):
        self.new_image_tag = new_image_tag
    def validate_new_image_tag(self, new_image_tag):
        existing_image = self.images.get_user_image_from_tag(
                                        self.requester,
                                        new_image_tag,
                                        expected=None)
        if self.image_tag == new_image_tag:
            # same name for the modified image.
            # this would overwrite the existing one.
            # we will let the user confirm this.
            self.images.warn_overwrite_image(self.requester, existing_image.fullname)
            return ImageShellSession.NAME_NEEDS_CONFIRM
        else:
            if existing_image:
                self.requester.stderr.write('Bad name: Image already exists.\n')
                return ImageShellSession.NAME_NOT_OK
        if not validate_image_tag(self.requester, new_image_tag):
            return ImageShellSession.NAME_NOT_OK
        return ImageShellSession.NAME_OK
    def finalize(self):
        if self.new_image_tag:
            image_fullname = '%s/walt-node:%s' % (
                    self.requester.username, self.new_image_tag)
            # with the walt image cp command, the client sends a request to start a
            # container for receiving, then immediately starts to send a tar archive,
            # and then tries to commit the container through rpyc commands.
            # if there is little data to transfer, the container might not be created yet,
            # because the initial request was not yet handled by the server...
            # so we have to ensure here that the container is created, and then wait for
            # its completion.
            while self.container_name not in self.docker.list_containers():
                time.sleep(0.05)
            self.docker.wait_container(self.container_name)
            self.docker.commit(self.container_name, image_fullname,
                    'Image modified using walt image shell')
            if self.image_tag == self.new_image_tag:
                # same name, we are modifying the image
                image = self.images.get_user_image_from_tag(self.requester, self.new_image_tag)
                # if image is mounted, umount/mount it in order to make
                # the nodes reboot with the new version
                if image.mounted:
                    # umount
                    self.images.umount_used_image(image)
                    # re-mount
                    self.images.update_image_mounts()
                # done.
                self.requester.stdout.write('Image %s updated.\n' % self.new_image_tag)
            else:
                # we are saving changes to a new image, leaving the initial one
                # unchanged
                self.images.register_image(image_fullname, True)
                self.requester.stdout.write('New image %s saved.\n' % self.new_image_tag)
        self.store.cleanup_session(self)
    def cleanup(self):
        self.docker.stop_container(self.container_name)





def node_exists(db, mac):
    return db.select_unique("nodes", mac=mac) != None

def register_node(  devices, db, \
                    mac, ip, node_type, \
                    image_fullname, **kwargs):
    # insert in table devices if missing
    if not db.select_unique("devices", mac=mac):
        devices.add(type=node_type, mac=mac, ip=ip)
    # insert in table nodes
    db.insert('nodes', mac=mac, image=image_fullname)
    db.commit()

def finalize_registration(current_requests, mac, **kwargs):
    current_requests.remove(mac)

def update_images_and_finalize(images, image_fullname, dhcpd, **kwargs):
    images.set_image_ready(image_fullname)
    # mount needed images
    # refresh the dhcpd conf
    images.update_image_mounts()
    dhcpd.update()
    # we are all done
    finalize_registration(**kwargs)

class AsyncPullTask(object):
    def __init__(self, docker, image_fullname, finalize_cb, finalize_cb_kwargs):
        self.docker = docker
        self.image_fullname = image_fullname
        self.finalize_cb = finalize_cb
        self.finalize_cb_kwargs = finalize_cb_kwargs
    def perform(self):
        # pull image: this is where things may take some time...
        self.docker.pull(self.image_fullname)
    def handle_result(self, res):
        # this should go fast
        self.finalize_cb(**self.finalize_cb_kwargs)

def handle_registration_request(
                db, docker, blocking, mac, images, node_type, \
                current_requests, **kwargs):
    if mac in current_requests or node_exists(db, mac):
        # this is a duplicate request, we already have registered
        # this node or it is being registered
        return
    current_requests.add(mac)
    image_fullname = images.get_default_image(node_type)
    image_is_new = image_fullname not in images
    # if image is new, register it before the node
    # (cf db integrity constraint)
    if image_is_new:
        images.register_image(image_fullname, False)
    # register the node
    full_kwargs = dict(
        db = db,
        images = images,
        mac = mac,
        node_type = node_type,
        image_fullname = image_fullname,
        current_requests = current_requests,
        **kwargs
    )
    register_node(**full_kwargs)
    # if image is new
    if image_is_new:
        # we have to pull an image, that will be long,
        # let's do this asynchronously
        blocking.do(AsyncPullTask(docker, image_fullname,
                            update_images_and_finalize, full_kwargs))
    else:
        finalize_registration(**full_kwargs)

class NodeRegistrationHandler(object):
    def __init__(self, blocking, sock, sock_file, **kwargs):
        self.sock_file = sock_file
        self.blocking = blocking
        self.mac = None
        self.ip = None
        self.node_type = None
        self.kwargs = kwargs

    # let the event loop know what we are reading on
    def fileno(self):
        return self.sock_file.fileno()
    # the node register itself in its early bootup phase,
    # thus the protocol is simple: based on text lines
    def readline(self):
        return self.sock_file.readline().strip()
    # when the event loop detects an event for us, we
    # know a log line should be read.
    def handle_event(self, ts):
        if self.mac == None:
            self.mac = self.readline()
        elif self.ip == None:
            self.ip = self.readline()
        elif self.node_type == None:
            self.node_type = self.readline()
            handle_registration_request(
                blocking = self.blocking,
                mac = self.mac,
                ip = self.ip,
                node_type = self.node_type,
                **self.kwargs
            )
            # tell the event_loop that we can be removed
            return False
    def close(self):
        self.sock_file.close()



from collections import defaultdict

class WaitInfo(object):
    def __init__(self):
        self.qid_to_macs = defaultdict(set)
        self.mac_to_qids = defaultdict(set)
        self.queues = {}

    def wait(self, requester, q, nodes):
        if nodes == None:
            q.put(0)    # unblock the client
        unreachable_nodes = [ node for node in nodes if node.reachable == 0 ]
        if len(unreachable_nodes) == 0:
            q.put(0)    # unblock the client
        # ok, the client will really have to wait
        qid = id(q)
        self.queues[qid] = q
        for node in unreachable_nodes:
            self.mac_to_qids[node.mac].add(qid)
            self.qid_to_macs[qid].add(node.mac)

    def node_bootup_event(self, node):
        print 'node bootup', node.name
        qids = self.mac_to_qids[node.mac]
        del self.mac_to_qids[node.mac]
        for qid in qids:
            self.qid_to_macs[qid].remove(node.mac)
            if len(self.qid_to_macs[qid]) == 0:
                # the queue is no more associated with any awaited nodes
                try:
                    self.queues[qid].put(0) # unblock the client
                except ReferenceError:
                    # the client already cancelled the wait
                    # thus the queue object no longer exists
                    # we can safely ignore that.
                    pass
                del self.queues[qid]
                del self.qid_to_macs[qid]



# coding=utf-8
import sys, curses, os, cPickle as pickle
from walt.common.evloop import EventLoop
from walt.common.tools import failsafe_mkfifo
from walt.server.const import UI_FIFO_PATH

class Console(object):
    def __init__(self):
        stdscr = curses.initscr()
        curses.curs_set(0) # invisible cursor
        max_y, max_x = stdscr.getmaxyx()
        stdscr.erase()
        stdscr.addstr('WalT server console', curses.A_BOLD | curses.A_UNDERLINE)
        self.status_win = stdscr.subwin(1, max_x, 2, 0)
        self.explain_win = stdscr.subwin(4, 0)
        self.BOX_DRAWING={
            u'┼': curses.ACS_PLUS,
            u'┌': curses.ACS_ULCORNER,
            u'─': curses.ACS_HLINE,
            u'│': curses.ACS_VLINE,
            u'┐': curses.ACS_URCORNER,
            u'┘': curses.ACS_LRCORNER,
            u'└': curses.ACS_LLCORNER
        }
        stdscr.refresh()
        failsafe_mkfifo(UI_FIFO_PATH)
        self.fifo = os.fdopen(
            os.open(UI_FIFO_PATH, os.O_RDWR | os.O_NONBLOCK), 'r', 0)
    def __del__(self):
        curses.endwin()
        self.fifo.close()
        os.remove(UI_FIFO_PATH)
    def set_win_text(self, win, text, flags, erase):
        if erase:
            win.erase()
        for c in text:
            if c in self.BOX_DRAWING:
                c = self.BOX_DRAWING[c]
            else:
                c = str(c)
            try:
                win.addch(c, flags)
            except Exception:
                raise Exception('Issue with %s in %s' % (repr(c), repr(text)))
        win.refresh()
    def handle_request(self, req, text):
        if req == 'STATUS':
            win = self.status_win
            flags = curses.A_BOLD
            text = 'Status: ' + text.strip('\n')
            erase = True
        elif req == 'EXPLAIN':
            win = self.explain_win
            flags = 0
            erase = True
        elif req == 'TODO':
            win = self.explain_win
            flags = curses.A_BOLD
            erase = False
        self.set_win_text(win, text, flags, erase)
    def run(self):
        ev_loop = EventLoop()
        ev_loop.register_listener(self)
        ev_loop.loop()
    def fileno(self):
        return self.fifo.fileno()
    def handle_event(self, ts):
        args = pickle.load(self.fifo)
        self.handle_request(*args)

def run():
    console = Console()
    try:
        console.run()
    except KeyboardInterrupt:
        del console
#    except Exception as e:
#        del console
#        raise e




#!/usr/bin/env python

import uuid
from walt.common.evloop import EventLoop
from walt.common.tcp import TCPServer
from walt.common.constants import WALT_SERVER_TCP_PORT
from walt.server.images.manager import NodeImageManager
from walt.server.devices.manager import DevicesManager
from walt.server.db import ServerDB
from walt.server.logs import LogsManager
from walt.server.network.dhcpd import DHCPServer
from walt.server.interactive import InteractionManager
from walt.server.transfer import TransferManager
from walt.server.blocking import BlockingTasksManager
from walt.server.nodes.manager import NodesManager
from walt.server.mydocker import DockerClient
from walt.server.tools import format_sentence_about_nodes

class Server(object):

    def __init__(self, ui):
        self.ui = ui
        self.ev_loop = EventLoop()
        self.db = ServerDB()
        self.docker = DockerClient()
        self.blocking = BlockingTasksManager()
        self.devices = DevicesManager(self.db)
        self.dhcpd = DHCPServer(self.db)
        self.images = NodeImageManager(self.db, self.blocking, self.dhcpd, self.docker)
        self.tcp_server = TCPServer(WALT_SERVER_TCP_PORT)
        self.logs = LogsManager(self.db, self.tcp_server, self.blocking)
        self.interaction = InteractionManager(\
                        self.tcp_server, self.ev_loop)
        self.transfer = TransferManager(\
                        self.tcp_server, self.ev_loop)
        self.nodes = NodesManager(  db = self.db,
                                    tcp_server = self.tcp_server,
                                    blocking = self.blocking,
                                    images = self.images.store,
                                    dhcpd = self.dhcpd,
                                    docker = self.docker,
                                    devices = self.devices)
        self.tcp_server.join_event_loop(self.ev_loop)
        self.blocking.join_event_loop(self.ev_loop)
        self.db.plan_auto_commit(self.ev_loop)

    def update(self):
        self.ui.task_start('Scanning walt devices and images...')
        self.ui.task_running()
        # ensure the dhcp server is running,
        # otherwise the switches may have ip addresses
        # outside the WalT network, and we will not be able
        # to communicate with them when trying to update
        # the topology.
        self.dhcpd.update(force=True)
        self.ui.task_running()
        # topology exploration
        self.devices.rescan(ui=self.ui)
        self.ui.task_running()
        # re-update dhcp with any new device discovered
        self.dhcpd.update()
        self.ui.task_running()
        # mount images needed
        self.images.update()
        self.ui.task_done()

    def cleanup(self):
        self.images.cleanup()
        self.blocking.cleanup()

    def set_image(self, requester, node_set, image_tag, warn_unknown_topology):
        nodes = self.nodes.parse_node_set(requester, node_set)
        if nodes == None:
            return # error already reported
        nodes_ok, nodes_ko = self.nodes.filter_on_connectivity( \
                            requester, nodes, warn_unknown_topology)
        macs = [ n.mac for n in nodes_ok ]
        if self.images.set_image(requester, macs, image_tag):
            requester.stdout.write(format_sentence_about_nodes(
                '%s will now boot ' + image_tag + '.' ,
                [n.name for n in nodes_ok]) + '\n')

    def rename_device(self, requester, old_name, new_name):
        self.devices.rename(requester, old_name, new_name)
        self.dhcpd.update()

    def device_rescan(self, requester):
        self.devices.rescan(requester=requester)
        self.dhcpd.update()

    def forget_device(self, device_name):
        self.db.forget_device(device_name)
        self.dhcpd.update()




import os, sys, cPickle as pickle
from walt.server.const import UI_FIFO_PATH

class UIManager(object):
    def ui_running(self):
        return os.path.exists(UI_FIFO_PATH)
    def request_ui_update(self, *args):
        if self.ui_running():
            with open(UI_FIFO_PATH, 'w') as fifo:
                pickle.dump(args, fifo)
            return True
        else:
            return False
    def task_start(self, msg, explain=None, todo=None):
        self.task_idx = 0
        self.task_msg = msg
        self.task_explain = explain
        self.task_todo = todo
        self.task_explained_ui = False
        self.set_status(msg)
    def task_running(self):
        if self.task_explain and not self.task_explained_ui:
            self.task_explained_ui = self.set_explain(
                self.task_explain,
                self.task_todo,
                ui_only = True
            )
        status_text = "%s %s" % \
                (self.task_msg, '|/-\\'[self.task_idx])
        self.set_status(status_text, ui_only=True)
        self.task_idx = (self.task_idx + 1)%4
    def task_done(self):
        status_text = "%s %s" % (self.task_msg, 'done')
        self.set_status(status_text)
        self.request_ui_update('EXPLAIN', '')
    def task_failed(self, error_msg):
        status_text = "%s %s" % (self.task_msg, 'FAILED!!')
        self.set_status(status_text)
        self.set_explain('ERROR:\n' + error_msg, None)
    def update_topic(self, topic, text, ui_only=False):
        if not ui_only:
            print '**', text
        return self.request_ui_update(topic, text)
    def set_status(self, *args, **kwargs):
        return self.update_topic('STATUS', *args, **kwargs)
    def set_explain(self, explain, todo, **kwargs):
        succeeded = self.update_topic('EXPLAIN', explain, **kwargs)
        if succeeded and todo is not None:
            succeeded = self.update_topic('TODO', todo, **kwargs)
        return succeeded



#!/usr/bin/env python

import rpyc, sys, datetime, cPickle as pickle
import walt.server as server
from walt.server.network import setup
from walt.server.tools import AutoCleaner
from walt.server.ui.manager import UIManager
from walt.common.daemon import WalTDaemon
from walt.common.constants import           \
                 WALT_SERVER_DAEMON_PORT,   \
                 WALT_NODE_DAEMON_PORT

WALT_SERVER_DAEMON_VERSION = 0.1

class ClientMirroringService(rpyc.Service):
    services_per_node = {}

    def on_connect(self):
        node_id = id(self._conn.root)
        self._node_id = node_id
        ClientMirroringService.services_per_node[node_id] = self

    def on_disconnect(self):
        del ClientMirroringService.services_per_node[self._node_id]

    def __del__(self):
        self._client = None

    def register_client(self, client):
        self._client = client

    @staticmethod
    def link_node_to_client(node, client):
        service = ClientMirroringService.services_per_node[id(node)]
        service.register_client(client)

    # forward all other method accesses to self._client
    def __getattr__(self, attr_name):
        return getattr(self._client, attr_name)

class ServerToNodeLink:
    def __init__(self, ip_address, client = None):
        self.node_ip = ip_address
        self.client = client

    def __enter__(self):
        if self.client:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT,
                            service = ClientMirroringService)
            node_service = self.conn.root
            ClientMirroringService.link_node_to_client(node_service, self.client)
        else:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT)
        return self.conn.root

    def __exit__(self, type, value, traceback):
        self.conn.close()

class PlatformService(rpyc.Service):
    ALIASES=("WalT_Platform",)

    def __init__(self, *args, **kwargs):
        rpyc.Service.__init__(self, *args, **kwargs)
        self.server = server.instance
        self.images = server.instance.images
        self.devices = server.instance.devices
        self.nodes = server.instance.nodes
        self.logs = server.instance.logs

    def on_connect(self):
        self._client = self._conn.root

    def on_disconnect(self):
        self._client = None

    def exposed_device_rescan(self):
        self.server.device_rescan(self._client)

    def exposed_device_tree(self):
        return self.devices.topology.tree()

    def exposed_device_show(self):
        return self.devices.topology.show()

    def exposed_show_nodes(self, show_all):
        return self.nodes.show(self._client, show_all)

    def exposed_get_reachable_nodes_ip(self, node_set):
        return self.nodes.get_reachable_nodes_ip(
                        self._client, node_set)

    def exposed_get_device_ip(self, device_name):
        return self.devices.get_device_ip(
                        self._client, device_name)

    def exposed_get_node_ip(self, node_name):
        return self.nodes.get_node_ip(
                        self._client, node_name)

    def exposed_blink(self, node_name, blink_status):
        node_ip = self.nodes.get_reachable_node_ip(
                        self._client, node_name)
        if node_ip == None:
            return False # error was already reported
        with ServerToNodeLink(node_ip, self._client) as node_service:
            node_service.blink(blink_status)
        return True

    def exposed_parse_set_of_nodes(self, node_set):
        nodes = self.nodes.parse_node_set(self._client, node_set)
        if nodes:
            return [n.name for n in nodes]

    def exposed_includes_nodes_not_owned(self, node_set, warn):
        return self.nodes.includes_nodes_not_owned(self._client, node_set, warn)

    def exposed_poweroff(self, node_set, warn_unknown_topology):
        return self.nodes.setpower(self._client, node_set, False, warn_unknown_topology)

    def exposed_poweron(self, node_set, warn_unknown_topology):
        return self.nodes.setpower(self._client, node_set, True, warn_unknown_topology)

    def exposed_validate_node_cp(self, src, dst):
        return self.nodes.validate_cp(self._client, src, dst)

    def exposed_wait_for_nodes(self, q, node_set):
        self.nodes.wait(self._client, q, node_set)

    def exposed_rename(self, old_name, new_name):
        self.server.rename_device(self._client, old_name, new_name)

    def exposed_has_image(self, image_tag):
        return self.images.has_image(self._client, image_tag)

    def exposed_set_image(self, node_set, image_tag, warn_unknown_topology):
        self.server.set_image(self._client, node_set, image_tag, warn_unknown_topology)

    def exposed_is_device_reachable(self, device_name):
        return self.devices.is_reachable(self._client, device_name)

    def exposed_count_logs(self, history, **kwargs):
        unpickled_history = (pickle.loads(e) if e else None for e in history)
        return self.server.db.count_logs(history = unpickled_history, **kwargs)

    def exposed_forget(self, device_name):
        self.server.forget_device(device_name)

    def exposed_fix_image_owner(self, other_user):
        return self.images.fix_owner(self._client, other_user)

    def exposed_search_images(self, q, keyword):
        self.images.search(self._client, q, keyword)

    def exposed_clone_image(self, q, clonable_link, force=False):
        self.images.clone(requester = self._client,
                          q = q,
                          clonable_link = clonable_link,
                          force = force)

    def exposed_show_images(self):
        return self.images.show(self._client.username)

    def exposed_create_image_shell_session(self, image_tag):
        return self.images.create_shell_session(self._client, image_tag)

    def exposed_remove_image(self, image_tag):
        self.images.remove(self._client, image_tag)

    def exposed_rename_image(self, image_tag, new_tag):
        self.images.rename(self._client, image_tag, new_tag)

    def exposed_duplicate_image(self, image_tag, new_tag):
        self.images.duplicate(self._client, image_tag, new_tag)

    def exposed_validate_image_cp(self, src, dst):
        return self.images.validate_cp(self._client, src, dst)

    def exposed_node_bootup_event(self):
        node_ip, node_port = self._conn._config['endpoints'][1]
        self.devices.node_bootup_event(node_ip)
        node_name = self.devices.get_name_from_ip(node_ip)
        self.nodes.node_bootup_event(node_name)

    def exposed_add_checkpoint(self, cp_name, pickled_date):
        date = None
        if pickled_date:
            date = pickle.loads(pickled_date)
        self.logs.add_checkpoint(self._client, cp_name, date)

    def exposed_remove_checkpoint(self, cp_name):
        self.logs.remove_checkpoint(self._client, cp_name)

    def exposed_list_checkpoints(self):
        self.logs.list_checkpoints(self._client)

    def exposed_get_pickled_time(self):
        return pickle.dumps(datetime.datetime.now())

    def exposed_get_pickled_checkpoint_time(self, cp_name):
        return self.logs.get_pickled_checkpoint_time(self._client, cp_name)

class WalTServerDaemon(WalTDaemon):
    """WalT (wireless testbed) server daemon."""
    VERSION = WALT_SERVER_DAEMON_VERSION

    def getParameters(self):
        return dict(
                service_cl = PlatformService,
                port = WALT_SERVER_DAEMON_PORT,
                ev_loop = server.instance.ev_loop)

    def init_end(self):
        server.instance.ui.set_status('Ready.')

def run():
    ui = UIManager()
    if setup.setup_needed(ui):
        setup.setup(ui)
    myserver = server.Server(ui)
    with AutoCleaner(myserver) as server.instance:
        myserver.update()
        WalTServerDaemon.run()

if __name__ == "__main__":
    run()




import requests, uuid
from walt.server.images.search import \
        Search, \
        LOCATION_WALT_SERVER, LOCATION_DOCKER_HUB, \
        LOCATION_LABEL, LOCATION_PER_LABEL

# About terminology: See comment about it in image.py.

MSG_IMAGE_NOT_REMOTE_BELONGS_TO_WS = """\
Invalid clonable image link.
Images of the form 'server:%s/<image_name>' already belong to
your working set and thus are not clonable.
"""
MSG_USE_FORCE = """\
If this is what you want, rerun with --force.
"""
MSG_INVALID_CLONABLE_LINK = """\
Invalid clonable image link. Format must be:
[server|hub]:<user>/<image_name>
(tip: walt image search [<keyword>])
"""

def parse_clonable_link(requester, clonable_link):
    bad = False
    parts = clonable_link.split(':')
    if len(parts) != 2:
        bad = True
    if not bad:
        location = parts[0]
        if location not in LOCATION_LABEL.values():
            bad = True
    if not bad:
        parts = parts[1].split('/')
        if len(parts) != 2:
            bad = True
    if not bad:
        return (LOCATION_PER_LABEL[location],
                parts[0],
                parts[1])
    if bad:
        requester.stderr.write(MSG_INVALID_CLONABLE_LINK)
        return None, None, None


def retag_image_to_requester_ws(docker, image_store, requester, source_fullname):
    source_tag = source_fullname.split(':')[1]
    dest_fullname = '%s/walt-node:%s' % (requester.username, source_tag)
    if dest_fullname in image_store:
        # an image with the target name exists...
        existing_dest_image = image_store[dest_fullname]
        # if image is mounted, umount/mount it in order to make
        # the nodes reboot with the new version
        need_mount_umount = existing_dest_image.mounted
        if need_mount_umount:
            # umount
            image_store.umount_used_image(existing_dest_image)
        # remove existing image
        docker.rmi(dest_fullname)
        # re-tag the source image
        docker.tag(source_fullname, dest_fullname)
        if need_mount_umount:
            # re-mount
            image_store.update_image_mounts()
    else:
        docker.tag(source_fullname, dest_fullname)
        image_store.register_image(dest_fullname, True)

def perform_clone(requester, docker, clonable_link, image_store, force):
    remote_location, remote_user, remote_tag = parse_clonable_link(
                                                requester, clonable_link)
    if remote_location == None:
        return # error already reported
    # actually we should validate the 3 components (user, tag and location), but
    # let's get a larger set of information (especially about other existing
    # images with the same tag, because we should take care not to overwrite
    # them)
    def validate(user, tag, location):
        return tag == remote_tag
    # search
    result = Search(docker, requester, 'Validating...').search(validate)
    # check that the requested image is in the resultset
    if      len(result) == 0 or \
            remote_user not in result[remote_tag] or \
            remote_location not in result[remote_tag][remote_user]:
        requester.stderr.write(
            'No such remote image. Use walt image search <keyword>.\n')
        return
    # filter out error cases
    users = result[remote_tag]
    locations = users[remote_user]
    would_overwrite_ws_image = False
    if requester.username == remote_user:
        if remote_location == LOCATION_WALT_SERVER:
            requester.stderr.write(
                MSG_IMAGE_NOT_REMOTE_BELONGS_TO_WS % requester.username)
            return
        else:
            if len(locations) == 2:
                would_overwrite_ws_image = True
    else:
        if requester.username in users and \
                LOCATION_WALT_SERVER in users[requester.username]:
            would_overwrite_ws_image = True
    if would_overwrite_ws_image and not force:
        ws_image_fullname = "%s/walt-node:%s" % (requester.username, remote_tag)
        image_store.warn_overwrite_image(requester, ws_image_fullname)
        requester.stderr.write(MSG_USE_FORCE)
        return
    # proceed
    remote_fullname = "%s/walt-node:%s" % (remote_user, remote_tag)
    if remote_location == LOCATION_DOCKER_HUB:
        # need to pull the image
        if LOCATION_WALT_SERVER in locations:
            # ---------------------
            # caution:
            # <u1> wants to pull hub:<u2>/<image> (and then store it as
            # server:<u1>/<image>) but server:<u2>/<image> also exists and its
            # tag would be overwritten by the docker pull operation.
            # In order to avoid that, we tag server:<u2>/<image> again
            # temporarily.
            # ---------------------
            # generate a temporary docker image name
            temp_repo = 'walt-temp-' + str(uuid.uuid4()).split('-')[0]
            temp_fullname = "%s/%s:%s" % (requester.username, temp_repo, remote_tag)
            # save server:<u2>/<image> by tagging it with the temp name
            docker.tag(remote_fullname, temp_fullname)
            # pull hub:<u2>/<image> (we get an updated version of server:<u2>/<image>)
            docker.pull(remote_fullname, requester.stdout)
            # tag the updated version of server:<u2>/<image> to server:<u1>/<image>
            # (this will make the image appear in <u1>'s working set)
            retag_image_to_requester_ws(docker, image_store, requester, remote_fullname)
            # restore the tag on server:<u2>/<image>
            docker.tag(temp_fullname, remote_fullname)
            # remove the temporary tag
            docker.untag(temp_fullname)
        else:
            # pull the image
            docker.pull(remote_fullname, requester.stdout)
            # re-tag the image with the requester username (make it appear in its WS)
            retag_image_to_requester_ws(docker, image_store, requester, remote_fullname)
            docker.untag(remote_fullname) # remove the old tag
    else:
        # tag an image of the server
        retag_image_to_requester_ws(docker, image_store, requester, remote_fullname)
    requester.stdout.write('Done.\n')

class CloneTask(object):
    def __init__(self, q, requester, **kwargs):
        self.response_q = q
        self.requester = requester
        self.kwargs = kwargs
    def perform(self):
        perform_clone(  requester = self.requester,
                        **self.kwargs)
    def handle_result(self, res):
        if isinstance(res, requests.exceptions.RequestException):
            self.requester.stderr.write(
                'Network connection to docker hub failed.\n')
            res = None
        elif isinstance(res, Exception):
            raise res   # unexpected
        self.response_q.put(res)

# this implements walt image clone
def clone(blocking, **kwargs):
    blocking.do(CloneTask(**kwargs))




#!/usr/bin/env python
# coding=utf-8
import time, snimpy, sys
from walt.server import snmp, status, network
from walt.server.network.tools import lldp_update
from waltvlanconf import configure_main_switch_if_needed
from walt.common.tools import do
from walt.server.mydocker import DockerClient

MSG_WAITING_COMPANION_SERVICES = 'Waiting for companion services to be ready...'
MSG_DETECTING_MAIN_SWITCH = 'Detecting the main switch...'
MSG_AUTO_CONFIGURING_MAIN_SWITCH = 'Auto-configuring the main switch...'
MSG_ERROR_RESET_THE_SWITCH_REBOOT = '''\
Could not communicate with the main switch. Try to reset it to factory settings.
The server will reboot and retry this bootup procedure in a few seconds.
'''
MSG_DHCP_REQUEST = "Requesting an IP for this WalT server..."
MSG_CHECK_DOCKER = "Checking internet connection to the Docker Hub..."

EXPLAIN_NETWORK = u'''\
The following network setup is required:

┌───────────────┐
│main switch    │   port 1: Externally managed VLAN with DHCP and internet access
│1 2 3 4 5 6 7 8│   port 2: (this) walt server
└┼─┼─┼─┼─┼─┼─┼─┼┘   ports 3 and more: managed walt network (walt nodes, cascaded switches)
 │ │
 │ └ walt server
 └ to internet
'''

TODO_DETECTING_MAIN_SWITCH = u'''
If this is not the case already, please connect the main switch as described in order to continue.
Note that it may take a few minutes to detect the main switch even if it is connected.
'''

TODO_DHCP_REQUEST = u'''
The server (mac address %(mac)s) is now sending DHCP requests on the externally managed VLAN (switch port 1).
The setup process will continue when a DHCP offer is obtained.
'''

TODO_CHECK_DOCKER = u'''
The server failed to access hub.docker.com. Please verify that the server is allowed to reach internet.
The setup process will retry this periodically until it succeeds.
'''

def main_switch_conf_callback(local_ip, switch_ip, ui):
    ui.task_running()
    configured = configure_main_switch_if_needed(
                        snmp.Proxy(str(switch_ip), vlan=True))
    return configured

def setup_needed(ui):
    # check if an ip was already assigned, which 
    # shows that setup was already done
    ui.task_start(MSG_WAITING_COMPANION_SERVICES)
    server_ip = str(network.tools.get_server_ip())
    res = None
    while True:
        try:
            snmp_local = snmp.Proxy('localhost', lldp=True)
            if server_ip in snmp_local.lldp.get_local_ips():
                res = False    # setup already done
            else:
                res = True     # setup needed
        except snimpy.snmp.SNMPNoSuchObject:
            ui.task_running()
            time.sleep(1)
        else:
            break
    ui.task_done()
    return res

def check_docker(ui):
    ui.task_start(  MSG_CHECK_DOCKER,
                    explain=EXPLAIN_NETWORK,
                    todo=TODO_CHECK_DOCKER)
    docker = DockerClient()
    while not docker.self_test():
        ui.task_running()
        time.sleep(1)
    ui.task_done()

def setup(ui):
    ui.task_start(  MSG_DETECTING_MAIN_SWITCH,
                    explain=EXPLAIN_NETWORK,
                    todo=TODO_DETECTING_MAIN_SWITCH)
    # retrieve info about the main switch by
    # using the local lldp & snmp daemons
    snmp_local = snmp.Proxy('localhost', lldp=True)
    while len(snmp_local.lldp.get_neighbors()) == 0:
        ui.task_running()
        lldp_update()
        time.sleep(1)
    ui.task_done()
    main_switch_info = snmp_local.lldp.get_neighbors().values()[0]
    print 'main switch:', main_switch_info
    # assign a temporary ip in order to communicate
    # with the main switch, and, if not done yet, set up
    # its vlan configuration
    ui.task_start(MSG_AUTO_CONFIGURING_MAIN_SWITCH)
    ui.task_running()
    switch_ip = network.tools.ip(main_switch_info['ip'])
    reached, configured = network.tools.assign_temp_ip_to_reach_neighbor(
                                switch_ip,
                                main_switch_conf_callback,
                                ui)
    if reached:
        ui.task_done()
    else:
        ui.task_failed(MSG_ERROR_RESET_THE_SWITCH_REBOOT)
        time.sleep(10)
        do('reboot')
        time.sleep(10)
        sys.exit()
    # set up the server static ip on eth0 (walt testbed network)
    network.tools.set_server_ip()
    # get a server ip on eth0.169 (external VLAN) using DHCP
    mac_addr = network.tools.get_mac_address('eth0.169')
    msg = MSG_DHCP_REQUEST
    explain = EXPLAIN_NETWORK
    todo = TODO_DHCP_REQUEST % dict(mac = mac_addr)
    network.tools.dhcp_wait_ip('eth0.169', ui, msg, explain, todo)
    # check connection to the docker hub
    check_docker(ui)
    return True     # ok





MSG_SAME_USER="""\
Invalid username. According to your walt.conf file, you are '%s'!
"""

MSG_MOUNTED="""\
Cannot proceed because some images of %s are mounted:
%s
"""

MSG_OVERWRITING="""\
Cannot proceed because the following images of %s would overwrite
those with the same name in your working set:
%s
"""

MSG_NO_SUCH_USER="""\
Connot find any images belonging to a user with name '%s'.
Make sure you typed it correctly.
"""

MSG_CHANGED_OWNER="""\
Image %s now belongs to you.
"""

def fix_owner(images, docker, requester, other_user):
    if requester.username == other_user:
        requester.stderr.write(MSG_SAME_USER % other_user)
        return
    mounted = set()
    candidates = set()
    for image in images.values():
        if image.user == other_user:
            if image.mounted:
                mounted.add(image.tag)
            else:
                candidates.add(image)
    if len(mounted) > 0:
        requester.stderr.write(MSG_MOUNTED % \
                (other_user, ', '.join(mounted)))
        return
    problematic = set()
    for image in candidates:
        if images.get_user_image_from_tag(requester, image.tag,
                                    expected = None, ready_only = False):
            problematic.add(image.tag)
    if len(problematic) > 0:
        requester.stderr.write(MSG_OVERWRITING % \
                (other_user, ', '.join(problematic)))
        return
    if len(candidates) == 0:
        requester.stderr.write(MSG_NO_SUCH_USER % other_user)
        return
    # ok, let's do it
    for image in candidates:
        # rename the docker image
        old_fullname = image.fullname
        new_fullname = "%s/walt-node:%s" % (requester.username, image.tag)
        docker.tag(old_fullname, new_fullname)
        docker.rmi(old_fullname)
        # update the store
        images.rename(old_fullname, new_fullname)
        requester.stdout.write(MSG_CHANGED_OWNER % image.tag)




from walt.server.images.image import parse_image_fullname
from walt.server.tools import \
        display_transient_label, hide_transient_label
from walt.server import const
from docker import Client
from datetime import datetime
import sys, requests, shlex, json

def docker_command_split(cmd):
    args = shlex.split(cmd)
    return dict(
        entrypoint=args[0],
        command=args[1:]
    )

class DockerClient(object):
    def __init__(self):
        self.c = Client(base_url='unix://var/run/docker.sock', version='auto', timeout=5)
    def self_test(self):
        try:
            self.c.search(term='walt-node')
        except:
            return False
        return True
    def pull(self, image_fullname, stdout = None):
        if stdout == None:
            stdout = sys.stdout
        fullname, name, repo, user, tag = parse_image_fullname(image_fullname)
        for idx, line in enumerate(
                self.c.pull(name, tag=requests.utils.quote(tag), stream=True)):
            progress = "\\|/-"[idx % 4]
            label = 'Downloading %s/%s... %s' % (user, tag, progress)
            display_transient_label(stdout, label)
        hide_transient_label(stdout, label)
        stdout.write('Downloading %s/%s... done.\n' % (user, tag))
    def tag(self, old_fullname, new_fullname):
        dummy1, new_name, dummy2, dummy3, new_tag = parse_image_fullname(new_fullname)
        self.c.tag(image=old_fullname, repository=new_name, tag=new_tag)
    def rmi(self, fullname):
        self.c.remove_image(image=fullname, force=True)
    def untag(self, fullname):
        self.rmi(fullname)
    def search(self, term):
        return self.c.search(term=term)
    def lookup_remote_tags(self, image_name):
        url = const.DOCKER_HUB_GET_TAGS_URL % dict(image_name = image_name)
        for elem in requests.get(url, timeout=3).json():
            tag = requests.utils.unquote(elem['name'])
            yield (image_name.split('/')[0], tag)
    def get_local_images(self):
        return sum([ i['RepoTags'] for i in self.c.images() ], [])
    def get_creation_time(self, image_fullname):
        for i in self.c.images():
            if image_fullname in i['RepoTags']:
                return datetime.fromtimestamp(i['Created'])
    def list_containers(self):
        return [ name.lstrip('/') for name in
                    sum([ cont['Names'] for cont in
                            self.c.containers(all=True) ], []) ]
    def start_container(self, image_fullname, cmd):
        params = dict(image=image_fullname)
        params.update(docker_command_split(cmd))
        cid = self.c.create_container(**params).get('Id')
        self.c.start(container=cid)
        return cid
    def wait_container(self, cid_or_cname):
        self.c.wait(container=cid_or_cname)
    def stop_container(self, cid_or_cname):
        try:
            self.c.kill(container=cid_or_cname)
        except:
            pass
        try:
            self.c.wait(container=cid_or_cname)
        except:
            pass
        try:
            self.c.remove_container(container=cid_or_cname)
        except:
            pass
    def commit(self, cid_or_cname, dest_fullname, msg):
        fullname, name, repo, user, tag = parse_image_fullname(dest_fullname)
        self.c.commit(
                container=cid_or_cname,
                repository=name,
                tag=tag,
                message=msg)
    def get_image_id(self, image_fullname):
        image_repo, image_tag = image_fullname.split(':')
        with open('/var/lib/docker/repositories-aufs') as conf_file:
            info = json.load(conf_file)
        return info['Repositories'][image_repo][image_tag]
    def get_image_layers(self, image_id):
        br = []
        while True:
            br.append('/var/lib/docker/aufs/diff/%s' % image_id)
            with open('/var/lib/docker/graph/%s/json' % image_id) as conf_file:
                info = json.load(conf_file)
            if 'parent' not in info:
                break
            else:
                image_id = info['parent']
        return br
